// Based on my shadow mapping demo based on a demo by Fabien Sanglard
// Reused for SSAO demo by Ingemar Ragnemalm.

// This demo demonstrates my own "SSAO hack", measuring concavities
// using pair of pixels on each side of the pixel being tested.

// This doesn't go all the way that it could. A more sophisticated version
// would scaled the measured area by distance.

// Also note: The SSAO effect is over-ephasized for demonstration purposes
// and needs to be tuned for other scenes.

// Originally written in C in some forgotten past. Coverted to C++ 2024.
// Only ested on my Mac at this time.

// uses framework Cocoa
// uses framework OpenGL
#define MAIN
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "MicroGlut.h"
#include "LoadTGA.h"
#include "VectorUtils4.h"
#include "GL_utilities.h"
#include "LittleOBJLoader.h"

// Expressed as float so gluPerspective division returns a float and not 0 (640/480 != 640.0/480.0).
#define RENDER_WIDTH 512.0
#define RENDER_HEIGHT 512.0

// We assign one texture unit in which to store the transformation.
#define TEX_UNIT 0
#define TEX_UNIT_DEPTH 1

//Camera position
vec3 p_camera = vec3(32,20,0);

//Camera lookAt
vec3 l_camera = vec3(2,0,-10);

// Use to activate/disable projTexShader
GLuint plainShaderId, ssaoShaderId;

FBOstruct *fbo;

#define NEAR 1.0
#define FAR 100.0
#define W 600
#define H 600

//-----------------------------matrices------------------------------
mat4	modelViewMatrix; // , textureMatrix;
mat4 projectionMatrix;

void loadShadowShader()
{
	plainShaderId = loadShaders("plain.vs", "plain.fs");
	glUseProgram(plainShaderId);
	glUniform1i(glGetUniformLocation(plainShaderId,"textureUnit"),TEX_UNIT);

	ssaoShaderId = loadShaders("ssao.vs", "ssao.fs");
	glUseProgram(ssaoShaderId);
	glUniform1i(glGetUniformLocation(ssaoShaderId,"texture1"),TEX_UNIT);
	glUniform1i(glGetUniformLocation(ssaoShaderId,"texture0"),TEX_UNIT_DEPTH);
}

// Arrays for ground and the Model references

GLfloat groundcolor[] = {0.3f,0.3f,0.3f,1};
GLfloat ground[] = {
							-35,2,-35,
							-35,2, 15,
							15,2, 15,
							15,2,-35};
GLfloat groundNormals[] = {
							0,1,0,
							0,1,0,
							0,1,0,
							0,1,0};
GLuint groundIndices[] = {0, 1, 2, 0, 2, 3};

GLfloat square[] = {
							-1,-1,0,
							-1,1, 0,
							1,1, 0,
							1,-1, 0};
GLuint squareIndices[] = {0, 1, 2, 0, 2, 3};

GLfloat cube[] = {
							-10.0f,   0.0f,  10.0f,
							 10.0f,   0.0f,  10.0f,
							 10.0f,  20.0f,  10.0f,
							-10.0f,  20.0f,  10.0f,
							
							-10.0f,   0.0f, -10.0f,
							-10.0f,  20.0f, -10.0f,
							 10.0f,  20.0f, -10.0f,
							 10.0f,   0.0f, -10.0f,

							-10.0f,  20.0f, -10.0f,
							-10.0f,  20.0f,  10.0f,
							 10.0f,  20.0f,  10.0f,
							 10.0f,  20.0f, -10.0f,

							-10.0f,   0.0f, -10.0f,
							 10.0f,   0.0f, -10.0f,
							 10.0f,   0.0f,  10.0f,
							-10.0f,   0.0f,  10.0f,

							 10.0f,   0.0f, -10.0f,
							 10.0f,  20.0f, -10.0f,
							 10.0f,  20.0f,  10.0f,
							 10.0f,   0.0f,  10.0f,

							-10.0f,   0.0f, -10.0f,
							-10.0f,   0.0f,  10.0f,
							-10.0f,  20.0f,  10.0f,
							-10.0f,  20.0f, -10.0f};
GLfloat cubeTex[] = {
							0.0f, 0.0f,
							1.0f, 0.0f,
							1.0f, 1.0f,
							0.0f, 1.0f,

							1.0f, 0.0f,
							1.0f, 1.0f,
							0.0f, 1.0f,
							0.0f, 0.0f,

							0.0f, 1.0f,
							0.0f, 0.0f,
							1.0f, 0.0f,
							1.0f, 1.0f,

							1.0f, 1.0f,
							0.0f, 1.0f,
							0.0f, 0.0f,
							1.0f, 0.0f,

							1.0f, 0.0f,
							1.0f, 1.0f,
							0.0f, 1.0f,
							0.0f, 0.0f,

							0.0f, 0.0f,
							1.0f, 0.0f,
							1.0f, 1.0f,
							0.0f, 1.0f};
GLuint cubeIndices[] = {
							0, 1, 2, 0, 2, 3, // Vilka mste flippas?
							4, 5, 6, 4, 6, 7,
							8, 9, 10, 8, 10, 11,
							12, 13, 14, 12, 14, 15,
							16, 17, 18, 16, 18, 19,
							20, 21, 22, 20, 22, 23};

Model *groundModel, *torusModel, *sphereModel, *squareModel, *cubeModel;

void loadObjects(void)
{
// The shader must be loaded before this is called!
	if (plainShaderId == 0)
		printf("Warning! Is the shader not loaded?\n");

	squareModel = LoadDataToModel(
			(vec3 *)square, NULL, NULL, NULL,
			squareIndices, 4, 6);
	
	groundModel = LoadDataToModel(
			(vec3 *)ground, (vec3 *)groundNormals, NULL, NULL,
			groundIndices, 4, 6);
	
	cubeModel = LoadDataToModel((vec3 *)cube, NULL, (vec2 *)cubeTex, NULL,
			cubeIndices, 4*6, 6*6); // , plainShaderId, "in_Position", "in_Normal", "in_TexCoord");

	torusModel = LoadModel("torus.obj"); // , plainShaderId, "in_Position", "in_Normal", "");
	sphereModel = LoadModel("stanford-bunny.obj"); // , plainShaderId, "in_Position", "in_Normal", "");

}

float a = 4.5;

void drawObjects(GLuint shader)
{
	mat4 mv2, tx2, trans, rot, scale;
	
	a += 0.01;
	
	glUniformMatrix4fv(glGetUniformLocation(plainShaderId, "projectionMatrix"), 1, GL_TRUE, projectionMatrix.m);

	// Ground
	glUniform1f(glGetUniformLocation(plainShaderId, "shade"), 0.7); // Dark ground
	glUniformMatrix4fv(glGetUniformLocation(plainShaderId, "modelViewMatrix"), 1, GL_TRUE, modelViewMatrix.m);
	DrawModel(groundModel, plainShaderId, "in_Position", "in_Normal", "");
	
	glUniform1f(glGetUniformLocation(plainShaderId, "shade"), 0.9); // Brighter objects
	
	// One torus
	trans = T(0,4,-16);
	scale = S(2.0, 2.0, 2.0);
	trans = trans * scale; // Apply on both
	mv2 = modelViewMatrix * trans; // Apply on both
	// Upload both!
	glUniformMatrix4fv(glGetUniformLocation(plainShaderId, "modelViewMatrix"), 1, GL_TRUE, mv2.m);
	DrawModel(torusModel, plainShaderId, "in_Position", "in_Normal", "");

	// The other torus
	trans = T(0,4,-16);
	scale = S(2.0, 2.0, 2.0);
	rot = Ry(3.14/2);
	trans = trans * rot; // Apply on both
	trans = trans * scale; // Apply on both
	mv2 = modelViewMatrix * trans; // Apply on both
	// Upload
	glUniformMatrix4fv(glGetUniformLocation(plainShaderId, "modelViewMatrix"), 1, GL_TRUE, mv2.m);
	DrawModel(torusModel, plainShaderId, "in_Position", "in_Normal", "");
	
	// The sphere (or rabbit)
	trans = T(0,0,-5);
	scale = S(50.0, 50.0, 50.0);
	trans = trans * scale * Ry(a);
	mv2 = modelViewMatrix * trans; // Apply on both
	// Upload
	glUniformMatrix4fv(glGetUniformLocation(plainShaderId, "modelViewMatrix"), 1, GL_TRUE, mv2.m);
	DrawModel(sphereModel, plainShaderId, "in_Position", "in_Normal", "");
}

void renderScene(void) 
{
	// Setup projection matrix
	projectionMatrix = perspective(45, RENDER_WIDTH/RENDER_HEIGHT, 1, 100);
	
	// Setup the modelview
	vec3 up = vec3(0,1,0);
	modelViewMatrix = lookAtv(p_camera, l_camera, up);

// 1. Render scene to FBO

	useFBO(fbo, NULL, NULL);
	glViewport(0,0,RENDER_WIDTH,RENDER_HEIGHT);
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE); 
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Using the simple shader
	glUseProgram(plainShaderId);

	drawObjects(plainShaderId);
	glFlush();
	
//2. Render from camera.

	useFBO(NULL, fbo, NULL);
	
	glViewport(0,0,RENDER_WIDTH,RENDER_HEIGHT);
	
	//Enabling color write (previously disabled for light POV z-buffer rendering)
	glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE); 
	
	// Clear previous frame values
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Using the projTex shader
// THIS SHOULD BE DONE BY useFBO!
// But then that has to handle depth buffer part
	glUseProgram(ssaoShaderId);
	glActiveTexture(GL_TEXTURE0 + TEX_UNIT);
	glBindTexture(GL_TEXTURE_2D,fbo->texid);
	glActiveTexture(GL_TEXTURE0 + TEX_UNIT_DEPTH);
	glBindTexture(GL_TEXTURE_2D,fbo->depth);
	
	glCullFace(GL_BACK);
	glDisable(GL_CULL_FACE);
	DrawModel(squareModel, ssaoShaderId, "in_Position", "in_Normal", "");
	
	glutSwapBuffers();
}

void processNormalKeys(unsigned char key, int x, int y)
{
	if (key == 27) 
		exit(0);
}

int main(int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowPosition(100,100);
	glutInitWindowSize(RENDER_WIDTH,RENDER_HEIGHT);
	glutInitContextVersion(3, 2);
	glutCreateWindow("SSAO demo");
	
	loadShadowShader();
	loadObjects();
	fbo = initFBO2(RENDER_WIDTH,RENDER_HEIGHT, 0, 1);
	
	glEnable(GL_DEPTH_TEST);
	glClearColor(0,0,0,1.0f);
	glEnable(GL_CULL_FACE);
	
	glutDisplayFunc(renderScene);
	glutRepeatingTimer(20);
	glutKeyboardFunc(processNormalKeys);
	
	glutMainLoop();
	exit(0);
}
